import yaml
from . import nlp
import os
from rich import inspect, reconfigure, get_console
from rich.console import Console
from contextlib import redirect_stdout
from textwrap import dedent
from .timeout import timeout


def eval_single(path, category=None, override_prompt=None, **kwargs):
    with open(path, 'r') as f:
        config = yaml.safe_load(f.read())
    try:
        prompt = nlp.prompts[config['context']
                             ] if override_prompt is None else nlp.prompts[override_prompt]
        context = nlp.Context(
            name=config['name'], prompt=prompt)
    except Exception as e:
        print(f'ERROR: Could not parse file {f.name}')
        print(e)
        inspect(config, docs=False, methods=False)
        return None, None

    # this is to short-circuit the evaluator
    # return {'result': np.random.choice([True, False], size=kwargs['n']), 'name': config['name']}, None

    # check if disabled
    disabled = False
    if 'categories' in config:
        if 'disabled' in config['categories']:
            disabled = True
        if category is not None and category not in config['categories']:
            disabled = True
    if disabled:
        return None, None

    dir = os.path.dirname(path)
    with open(os.path.join(dir, config['prompt']), 'r') as f:
        query = f.read()
    # TODO Better way?
    # Make it so we can only get one function
    if context.prompt.stop is None:
        context.prompt.stop = ['def']
    elif 'def' not in context.prompt.stop:
        context.prompt.stop = ['def'] + context.prompt.stop
    context = nlp.code_completion(query, context, **kwargs)

    if not 'test' in config or config['test'] is None:
        result = {'name': config['name'], 'context': context, 'result': None}
        return result, ''

    with open(os.path.join(dir, config['test']), 'r') as f:
        test_code = f.read()
    runs = []
    exceptions = []
    dir_string = f"_FILE_DIR_='{dir}'\n"
    markdown = f'\n#### Query\n\n```py\n{context.query}\n\n```'
    for i, r in enumerate(context.responses):
        g = {}
        success = True
        try:
            with redirect_stdout(None):
                with timeout(seconds=10):
                    exec(dir_string + r + '\n' + test_code, g)
            if 'result' not in g:
                exceptions.append(
                    f'\nYou must have variable `result` defined. \n')
                success = False
            runs.append(g['result'])
        except Exception as e:
            exceptions.append(
                f'{str(e)}')
            runs.append(False)
            success = False
        markdown += f'\n#### Run {i}\n\n'
        code_str = r + '\n' + test_code
        markdown += f'```py\n{code_str}\n```\n'
        markdown += f'Output:\n```\n{"Success" if success else exceptions[-1]}\n```\n\n'
    result = {'name': config['name'], 'context': context, 'result': runs}

    # can be too long
    return result, markdown



import nlcc.prompts
import os
import openai
import unicodedata
from dataclasses import dataclass
from importlib_resources import files
import yaml


@dataclass
class Prompt:
    text: str = ''
    comment: str = '# '
    multiline_comment: str = None
    stop: list = None
    language: str = None
    emoji: str = '💻'


@dataclass
class Context:
    name: str = ''
    prompt: Prompt = None
    text: str = ''
    T: float = 0.0
    responses: tuple = None
    query_type: str = None


# load prompts on import
prompts = {}

for f in files(nlcc.prompts).iterdir():
    c = yaml.safe_load(f.read_text())
    for k, v in c.items():
        prompts.update({k: Prompt(**v)})


def guess_query_type(query):
    if len(query) == 0:
        return 'continue'
    if len(query.split('\n')) == 1:
        return 'comment'
    return 'code'


def code_completion(query, context, engine, query_type=None, T=0.0, n=1):
    context.T = T
    if len(context.text) == 0:
        context.text = context.prompt.text
    # count newlines to guess if we should insert comments (and how)
    if query_type is None:
        query_type = guess_query_type(query)
    if query_type == 'continue':
        query = context.text + query
    elif query_type == 'comment':
        if context.prompt.multiline_comment:
            query = context.prompt.multiline_comment + '\n' + \
                query + '\n' + context.prompt.multiline_comment + '\n'
        else:
            query = context.prompt.comment + ' ' + query + '\n'
        query = context.text + '\n' + \
            query if len(context.text) > 0 else query
    elif query_type == 'code':
        query = context.text + '\n' + \
            query if len(context.text) > 0 else query
    context.query = query
    context.query_type = query_type
    r = engine(query, T=T, stop=context.prompt.stop, n=n)
    context.responses = tuple([query + ri for ri in r])
    return context


def guess_context(query, engine, T=0.4):
    # check for exact match
    matched = None
    for p in prompts:
        if query.lower() == p.lower():
            matched = p
            e = prompts[p].emoji
    # try engine match
    if not matched:
        r, _ = engine(query, T=T)
        r = r.split()[0]
        e, r = r.split(',')
        if r in prompts:
            matched = r
    if matched:
        context = Context(e + ' ' + matched,
                          prompts[matched], prompts[matched].text)
    else:
        context = Context(e, Prompt())
    return context



import openai
import os
import time
from pyrate_limiter import Duration, Limiter, RequestRate

limiter = Limiter(RequestRate(23, Duration.MINUTE))


openai.api_key = os.getenv('OPENAI_API_KEY')


def code_engine(query, T=0.00, stop=None, n=1, max_tokens=896):
    while True:
        try:
            with limiter.ratelimit('codex', delay=True):
                response = openai.Completion.create(
                    engine="davinci-codex",
                    prompt=query,
                    temperature=T,
                    max_tokens=max_tokens,
                    top_p=1,
                    frequency_penalty=0.0,
                    presence_penalty=0,
                    stop=stop,
                    n=n,
                )
                break
        except openai.error.RateLimitError:
            print('Rate limit exceeded (even though we limit!), retrying...')
            time.sleep(15)
            continue

    return [r['text'] for r in response['choices']]


def nlp_engine(query, T=0.3):
    response = openai.Completion.create(
        engine="davinci",
        prompt=f"This takes text and identifies which computational chemistry library it is about and gives an appropriate emoji. \n\nText: \"Let's do something with mdtraj\"\nLibrary: ⚗,mdtraj\n###\nText: \"MD in gromacs\"\nLibrary: 🍏,gromacs\n###\nText: \"Protein analysis\"\nLibrary: 🌳,biopython\n###\nText: \"Write a slurm script\"\nLibrary: 📝,slurm\n###\nText: \"{query}\"\nLibrary:",
        temperature=T,
        max_tokens=60,
        top_p=1,
        best_of=3,
        frequency_penalty=0,
        presence_penalty=0,
        stop=["###"]
    )

    return response['choices'][0]['text'], response



from prompt_toolkit import PromptSession
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory
from prompt_toolkit.history import InMemoryHistory
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit import print_formatted_text
from enum import Enum


class Modes(Enum):
    QUERY = 0
    TEMPERATURE = 1
    SELECT_NRESPONSE = 2
    WRITE_FILE = 3
    READ_FILE = 4
    SELECT_CONTEXT = 5
    SELECT_RESPONSE = 6
    MULTILINE = 7

class PromptManager:
    def __init__(self):
        self.input_mode = [Modes.SELECT_CONTEXT]
        self.cli_prompt = ['👋']
    def push(self, prompt, mode):
        self.input_mode.append(mode)
        self.cli_prompt.append(prompt)
    def peek_mode(self):
        return self.input_mode[-1]
    def peek_cli_prompt(self):
        return self.cli_prompt[-1]
    def pop(self):
        self.cli_prompt.pop(), self.input_mode.pop()
    def __len__(self):
        return len(self.input_mode)

def text_iter(pm, extra_kbs):

    kb = KeyBindings()
    session = None

    @kb.add('enter')
    def _(event):
        if pm.peek_mode() != Modes.MULTILINE:
            event.current_buffer.validate_and_handle()
        else:
            event.current_buffer.insert_text('\n')


    @kb.add('c-down')
    def _(event):
        if pm.peek_mode() != Modes.MULTILINE:
            pm.push('multiline📜', Modes.MULTILINE)
        else:
            pm.pop()
    @kb.add('escape', 'enter')
    def _(event):
        if pm.peek_mode() == Modes.MULTILINE:
            pm.pop()
            event.current_buffer.validate_and_handle()




    history = InMemoryHistory()
    session = PromptSession(
        history=history,
        auto_suggest=AutoSuggestFromHistory(),
        enable_history_search=True,
        complete_while_typing=True,
        multiline=True,
        rprompt=lambda :f'{pm.peek_mode()}'
        #key_bindings=kb
    )

    def kb_wrapper(fxn):
        def kbf(e):
            fxn(e)
            print_formatted_text('')
        return kbf


    session.key_bindings = kb
    for k, v in extra_kbs.items():
        kb.add(k)(kb_wrapper(v))


    while True:
        q = session.prompt(lambda: f'nlcc|{pm.peek_cli_prompt()}:>')
        yield q



from .eval import eval_single
from nlcc.openai import code_engine, nlp_engine
import os
from rich.markdown import Markdown
import pyperclip
from rich.syntax import Syntax
from rich.console import Console
from rich import pretty
from rich import inspect
from .nlp import Context, Prompt, guess_context, code_completion
from . import nlp
import click
from importlib_metadata import metadata
from .prompt import text_iter, Modes, PromptManager
from textwrap import dedent

pretty.install()


def process_n_response(query, console):
    try:
        new_n_responses = int(query)
    except ValueError as e:
        console.print(
            "\tFailed at setting n_responses with value:", query)
        return None
    if 1 > new_n_responses:
        console.print(
            "\tFailed at setting n_responses to value:", new_n_responses)
        console.print("\tn_responses must be an integer >= 1")
        return None
    return new_n_responses


def process_temperature(query, code_temp, nlp_temp, console):
    try:
        new_nlp_temp, new_code_temp = [float(s) for s in query.split(',')]
    except (ValueError, IndexError) as e:
        console.print(
            "\tFailed at setting temperature with command:", query)
        return
    if 0 > new_code_temp or new_code_temp > 1:
        console.print(
            "\tFailed at setting code temperature with command:", query)
        console.print("\tTemperature out of range [0,1]")
        return
    else:
        code_temp = new_code_temp
        console.print("\tSetting code🧊 query temperature to", code_temp)

    if 0 > new_code_temp or new_code_temp > 1:
        console.print(
            "\tFailed at setting GPT3 temperature with command:", query)
        console.print("\tTemperature out of range [0,1]")
        return
    else:
        nlp_temp = new_nlp_temp
        console.print("\tSetting GPT3🔥 query temperature to", nlp_temp)
    return code_temp, nlp_temp


def get_engine(engine, console=None):
    if engine == 'openai':
        if console:
            console.print('Using OpenAI Engine💰💰💰')
        from .openai import nlp_engine, code_engine
    elif engine == 'huggingface':
        if console:
            console.print('Using Huggingface Engine🤗🤗🤗')
        from .huggingface import nlp_engine
    else:
        if console:
            console.print('Unkown engine', engine)
        exit(1)
    return nlp_engine, code_engine


_DEFAULT_NLP_T = 0.3
_DEFAULT_CODE_T = 0.0


@click.command()
@click.argument('input_file', default=None, required=False)
@click.option('--n_responses', default=1)
@click.option('--engine', default='openai')
@click.option('--help', is_flag=True)
def main(input_file, engine, help, n_responses):
    console = Console()
    context = Context("", Prompt())
    pm = PromptManager()
    code_temp, nlp_temp = _DEFAULT_CODE_T, _DEFAULT_NLP_T

    nlp_engine, code_engine = get_engine(engine, console)

    def query_file_text(input_file, context):
        if not os.path.exists(input_file):
            console.print(f"Input file not found: {input_file}")
            return None
        query = open(input_file, 'r').read()
        context = code_completion(
            query, context, code_engine, T=code_temp, n=n_responses)
        response_text = context.responses
        for idx, response in enumerate(response_text):
            if len(response_text) > 1:
                console.print(f"## Option {idx+1}")
            console.print(Syntax(query, context.prompt.language,
                                 theme='monokai', line_numbers=False))
            console.print(Syntax(response, context.prompt.language,
                                 theme='monokai', line_numbers=False))
        return context

    if input_file is not None:
        context = Context("", Prompt())
        query_file_text(input_file, context)
        exit()

    # welcome message
    readme = metadata('nlcc')['Description']
    kb_info = readme.split('## key bindings')[-1]
    help_message = '# 🧠 nlcc 🧠 \n' + '## Info on keybinds\n' + kb_info

    console.print(Markdown(help_message))
    if(help):
        exit()

    def help(e):
        console.print()
        console.print(Markdown(help_message))

    # make it here to have access to context
    def make_copy(e):
        if context is not None:
            pyperclip.copy(context.text)
            console.print('\n✨copied✨')
        else:
            console.print('\n🤔 nothing to copy')

    def status(e):
        console.print()
        console.print(Markdown(
            f'# 🧠 nlcc Status 🧠\n - nlp🔥: `T = {nlp_temp}` \n- code🧊: `T = {code_temp}` \n- n_responses🤏: `N = {n_responses}` '))
        inspect(context, title='Context',  docs=False)
        if len(pm) > 1:
            inspect(pm, title='Input Mode', docs=False)

    def reset_context(e):
        nonlocal context
        context = Context("", Prompt())
        pm.pop()
        pm.push('👋', Modes.SELECT_CONTEXT)

    def responses(e):
        pm.push('num_responses🥳', Modes.SELECT_NRESPONSE)

    def temperature(e):
        pm.push('🔥🧊', Modes.TEMPERATURE)
        console.print(
            f'\nEnter as 0.7,0.2. Currently nlp🔥:{nlp_temp} code🧊: {code_temp}', end='')

    def execute(e):
        g = {}
        console.print()
        exec(context.text, g)

    def write(e):
        pm.push('filename✍️📝', Modes.WRITE_FILE)

    def read(e):
        pm.push('filename👀📝', Modes.READ_FILE)

    kbs = {'c-w': make_copy, 'c-o': reset_context, 'c-z': execute, 'c-n': responses,
           'c-q': help, 'c-u': status, 'c-t': temperature, 'c-x': write, 'c-l': read}
    for i, query in enumerate(text_iter(pm, kbs)):

        if query.lower() == 'exit' or query.lower() == 'q' or query.lower() == 'quit':
            break
        elif pm.peek_mode() == Modes.SELECT_NRESPONSE:
            new_n = process_n_response(query, console)
            if new_n is not None:
                n_responses = new_n
        elif pm.peek_mode() == Modes.TEMPERATURE:
            new_t = process_temperature(
                query, code_temp, nlp_temp, console)
            if new_t is not None:
                code_temp, nlp_temp = new_t
        elif pm.peek_mode() == Modes.SELECT_CONTEXT:
            context = guess_context(query, nlp_engine, nlp_temp)
        elif pm.peek_mode() == Modes.READ_FILE:
            context_returned = query_file_text(query, context)
            if type(context_returned) == type(context):
                context = context_returned
        elif pm.peek_mode() == Modes.WRITE_FILE:
            if context:
                if query and os.path.normpath('(path-to-wiki)/foo/bar.txt').startswith('(path-to-wiki)'):
                    with open(query, 'w') as f:
                        f.write(context.text + '\n')
                        console.print(
                            f'✨wrote to {query}✨')
                else:
                    console.print(f'🤔 Not sure about this filepath: {query}')
            else:
                console.print('🤔 nothing to write')
        elif pm.peek_mode() == Modes.SELECT_RESPONSE:
            try:
                i = int(query) - 1
                context.text = context.responses[i]
                console.print(Syntax(context.text, context.prompt.language,
                                     theme='monokai', line_numbers=False))
                context.responses = None
            except Exception as e:
                console.print(e)
                console.print('🤮 what was that? Please pick response')
                continue
        else:
            context = code_completion(
                query, context, code_engine, T=code_temp, n=n_responses)
            if len(context.responses) == 1:
                context.text = context.responses[0]
                console.print(Syntax(context.text, context.prompt.language,
                                     theme='monokai', line_numbers=False))
                context.responses = None
            else:
                for ridx, r in enumerate(context.responses):
                    console.print(f"## Option {ridx+1}")
                    console.print(Syntax(r, context.prompt.language,
                                         theme='monokai', line_numbers=False))
                pm.push('🤔which response', Modes.SELECT_RESPONSE)
                # do not want pop
                continue

        # reset mode
        pm.pop()
        if len(pm) == 0:
            pm.push(context.name if len(context.name) >
                    1 else 'context-free', Modes.QUERY)


@click.command()
@click.argument('yaml-files', type=click.Path(exists=True), nargs=-1)
@click.option('--out-dir', type=click.Path(exists=True))
@click.option('--n', default=1, help='number of respones')
@click.option('--engine', default='openai')
@click.option('--temperature', default=0.2)
def human_check(yaml_files, n, engine, temperature, out_dir=None):
    from datetime import date
    import yaml
    nlp_engine, code_engine = get_engine(engine)
    dict_list = []
    date_label = date.today().strftime("%d%b%Y")

    print('## Human Evaluations\n\n Follow links to add evaluation to evaluate\n\n')
    for y in yaml_files:
        if out_dir is None:
            out_dir = os.path.dirname(y)

        report, info = eval_single(
            y, engine=code_engine, n=n, T=temperature, category='human')
        if report is None:
            continue
        context = report["context"]
        print(f'### {report["name"]}\n\n')
        for ridx, r in enumerate(context.responses):
            result_dict = {}
            label = f"result_{report['name']}_d{date_label}_T{context.T}_r{ridx}"
            result_dict['label'] = label
            #result_dict['context'] = context
            result_dict['response_id'] = ridx
            result_dict['code_response'] = r
            result_dict['prompt'] = context.text
            result_dict['name'] = report['name'] + ' ' + str(ridx)
            result_dict['query'] = context.query
            result_dict['query_type'] = context.query_type
            result_dict['computer_result'] = report['result']
            if result_dict['computer_result'] is not None:
                result_dict['computer_result'] = result_dict['computer_result'][ridx]
            out_file = os.path.join(out_dir, label)+'.yml'
            yaml.dump(result_dict, open(out_file, 'w'))
            print(f'* [Reponse {ridx}](human/{label}.html)')
        print('\n')


@click.command()
@click.argument('yaml-files', type=click.Path(exists=True), nargs=-1)
@click.option('--n', default=1, help='number of respones')
@click.option('--engine', default='openai')
@click.option('--terminal', default=False, is_flag=True)
@click.option('--temperature', default=0.2)
def eval(yaml_files, n, engine, temperature, terminal):
    if terminal is True:
        console = Console()
    from tabulate import tabulate
    nlp_engine, code_engine = get_engine(engine)
    table = []
    collapsables = []
    for y in yaml_files:
        report, info = eval_single(
            y, engine=code_engine, n=n, T=temperature, category='code')
        if report is None:
            continue
        if terminal is True:
            context = report["context"]
            console.print(Syntax(info, context.prompt.language,
                                 theme='monokai', line_numbers=False))
            for ridx, r in enumerate(context.responses):
                console.print(f"## Option {ridx+1}")
                console.print(Syntax(r, context.prompt.language,
                                     theme='monokai', line_numbers=False))
            console.print(report['name'])
            console.print(report['result'])
            continue
        if report is None:
            exit(1)
        if report['result'] is None:
            continue
        collapsables.append(
            f'<details>\n<summary>{report["name"]}</summary>\n{info}</details>')
        table.append([report['name']] +
                     ['Pass' if r else 'Fail' for r in report['result']])
    if terminal is True:
        return
    print('## Test Report')
    print('### Global Parameters')
    print('* Engine = ', engine)
    print('* n = ', n, '\n')
    print('### Code Results\n')
    print(tabulate(table, ['Test'] +
                   [f'Run {i}' for i in range(n)], tablefmt="github"))
    print('### Details\n')
    print('\n'.join(collapsables))


@click.command()
@click.argument('yaml-files', type=click.Path(exists=True), nargs=-1)
@click.argument('output', type=click.Path(exists=False))
@click.option('--n', default=1, help='number of respones')
@click.option('--engine', default='openai')
@click.option('--temperature', default=0.2)
@click.option('--prompt', default=None)
def benchmark(yaml_files, output, n, engine, temperature, prompt):
    from tabulate import tabulate
    nlp_engine, code_engine = get_engine(engine)
    table = []
    collapsables = []
    for y in yaml_files:
        print(y)
        report, _ = eval_single(
            y, engine=code_engine, n=n, T=temperature, override_prompt=prompt)
        if report['result'] is None:
            # disabled
            continue
        for r in report['result']:
            table.append([report['name']] +
                         [1 if r else 0] + [temperature])
    with open(output, 'w') as f:
        f.write(tabulate(table, ['name', 'result',
                                 'temperature'], tablefmt="plain"))
        f.write('\n')



import os
from glob import glob
from setuptools import setup

exec(open('nlcc/version.py').read())

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(name='nlcc',
      version=__version__,
      description='Natural language computational chemistry',
      author='Andrew & Frank',
      author_email='andrew.white@rochester.edu',
      url='https://github.com/whitead/nlcc',
      license='MIT',
      packages=['nlcc'],
      install_requires=[
          'click', 'importlib_resources',
          'importlib-metadata',
          'numpy', 'pyyaml~=5.0',
          'openai', 'pyperclip',
          'prompt_toolkit', 'tabulate', 'pyrate-limiter',
          'rich'],
      test_suite='tests',
      zip_safe=True,
      entry_points='''
        [console_scripts]
        nlcc=nlcc.main:main
        nlcc-eval=nlcc.main:eval
        nlcc-bench=nlcc.main:benchmark
        nlcc-prompts=nlcc.main:prompts
        nlcc-human=nlcc.main:human_check
            ''',
      package_data={'nlcc': ['prompts/*yml']},
      long_description=long_description,
      long_description_content_type="text/markdown",
      classifiers=[
          "Programming Language :: Python :: 3",
          "License :: OSI Approved :: MIT License",
          "Operating System :: OS Independent",
          "Topic :: Scientific/Engineering :: Bio-Informatics",
          "Topic :: Scientific/Engineering :: Chemistry",
          "Topic :: Scientific/Engineering :: Artificial Intelligence"
      ]
      )
